# ** header v3.0
# This file is a part of the CaosDB Project.
#
# Copyright (C) 2021 Indiscale GmbH <info@indiscale.com>
# Copyright (C) 2021 Florian Spreckelsen <f.spreckelsen@indiscale.com>
# Copyright (C) 2021 Alexander Kreft
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see
# <https://www.gnu.org/licenses/>.
#
# ** end header
#

module CaosDB

using Libdl

"""
Chose the name of the library according to the OS you're running.
"""
library_name = (@static Sys.iswindows() ? "ccaosdb" : "libccaosdb")
if isempty(find_library(library_name))
    @error "Could not find $library_name"
end

module Exceptions

export evaluate_return_code, CaosDBException, GenericCaosDBException, CaosDBMessage

using Logging
using CaosDB

"""
The parent type of all CaosDB errors that can also be used for testing.
"""
abstract type CaosDBException <: Exception end

"""
A generic exception that will be raised in case of non-zero return
values of the calls to libccaosdb. May carry a message string and a
code.
"""
struct GenericCaosDBException <: CaosDBException
    msg::String
    code::Cint
end

Base.showerror(io::IO, e::CaosDBException) =
    print(io, "CaosDBException: ", e.msg, " (Status code ", e.code, ")")

"""
Struct containing Messages and codes for status codes<0 that do not
correspond to errors or success.
"""
struct CaosDBMessage
    msg::String
    code::Cint
end

Base.show(io::IO, m::CaosDBMessage) = print(io, m.msg, " (Status code ", m.code, ")")

"""
    function evaluate_return_code(code::Cint)

Evaluate the return code of a libccaosdb ccall and raise a
`GenericCaosDBException` in case of a non-zero return code.
"""
function evaluate_return_code(code::Cint)
    if code != 0
        msg = ccall(
            (:caosdb_get_status_description, CaosDB.library_name),
            Cstring,
            (Cint,),
            code,
        )
        if code > 0
            throw(GenericCaosDBException(unsafe_string(msg), code))
        else
            @info CaosDBMessage(unsafe_string(msg), code)
        end
    end
end

end # Exceptions

module Info

"""
Struct containing version information of the CaosDB server. Meant
mainly for internal usage; use `CaosDB.Connection.get_version_info` or
`CaosDB.Connection.print_version_info` to retrieve the version of the
connected CaosDB server.
"""
mutable struct _VersionInfo

    major::Cint
    minor::Cint
    patch::Cint
    pre_release::Cstring
    build::Cstring

    _VersionInfo() = new()

end

end # Info

module Utility

export get_env_var

using ..CaosDB

"""
    get_env_var(var[, default])

Return the environmental variable `var` if it exists, `default`
otherwise. If no `default` is given an empty string is returned
instead.
"""
function get_env_var(var::AbstractString, default::AbstractString = "")

    ret = ccall(
        (:caosdb_utility_get_env_var, CaosDB.library_name),
        Cstring,
        (Cstring, Cstring),
        var,
        default,
    )

    return unsafe_string(ret)

end

end # Utility

module Authentication

using ..CaosDB

"""
Struct containing a pointer to the wrapped cpp authenticator
class. Meant for internal use; call a
`CaosDB.Authentication.create_<authenticator>` function to create an
authenticator object from a configuration.
"""
mutable struct _Authenticator
    wrapped_authenticator::Ptr{Cvoid}

    function _Authenticator(managed_by_julia::Bool = false)
        auth = new()
        if managed_by_julia
            # Only append a finalizer for this if the object is
            # actually managed by Julia and not created and destroyed
            # internally by libcaosdb.
            function f(t)
                ccall(
                    (:caosdb_authentication_delete_authenticator, CaosDB.library_name),
                    Cint,
                    (Ref{_Authenticator},),
                    Ref{_Authenticator}(t),
                )
            end
            finalizer(f, auth)
        end
        return auth
    end
end

"""
    create_plain_password_authenticator(
        username::AbstractString,
        password::AbstractString,
    )

Return an authenticator object that contains a wrapped cpp
plain-password authenticator configured with `username` and
`password`.
"""
function create_plain_password_authenticator(
    username::AbstractString,
    password::AbstractString,
)

    auth = Ref{_Authenticator}(_Authenticator(true))

    err_code = ccall(
        (:caosdb_authentication_create_plain_password_authenticator, CaosDB.library_name),
        Cint,
        (Ref{_Authenticator}, Cstring, Cstring),
        auth,
        username,
        password,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return auth

end

end # Authentication

include("Connection.jl")

module Entity end

include("Transaction.jl")

end # CaosDB
